
/* Config: fake processing DOS 6 config.sys. This could be used in an 
autoexec.bat for a pre-dos 6 version to provide a menu capability for
selecting various configurations. Really this only allows us to pull in 
various sets of environment settings and choose a shell       

We do not impose size limits such of maximum number of tag names, length of
string arguments, etc., so this could easily be adapted to generate generic
menu driven apps based on the config.sys file syntax

*/  

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "parse.h"
#include "menu.h"
#include <alloc.h>


#define CONFIG_SYS "C:\\CONFIG.SYS"
#define COMSPEC "Z:\\COMMAND.COM" 
#define VERSION "1.0"
#define USAGE "config [-v -h -p] [config.sys file]"
#define HELP1 "Emulate DOS 6 config file processing"
#define HELP2 "-v: print version number and exit"
#define HELP3 "-h: print this helpful message and exit"
#define HELP4 "-p: parse config file only. Give short report"
#define HELP5 "default config.sys file is C:\\config.sys"

static char progname[FILENAME_MAX];
static char filename[FILENAME_MAX];
static char *shell_command_line;
static int parse_status = MY_NORMAL;

/* A tag is something like [COMMON]. As we process config.sys, when there is
a current tag we add set ...  lines to env and over-ride shell ... lines. We 
also process include lines and prepare for a menu, if needed */

struct tag_struct {
        int n_env;  
        char *name;
        char **env;  /* to hold pointers to SET = strings */
        char *shell; /* the current shell command line */
        struct menuitem_struct **my_menu;
        int nmenuitems;
        int menudefault;
        int menutimeout;
        int menuforeground;
        int menubackground;
} *my_tags;    

static int ntags;
static int max_env;

void forceuc(char *);
int do_tag(int);
#ifdef NO_STRDUP
char *strdup(char *);
#endif


int main(int argc, char **argv){

        int n,i,j,k,j_t=0;  /* j_t is current tag number */
        FILE *input;
        char c, *p;
        char *tokenbuf;
        char *current_tag;
        int choice;
        int active = 1;   /* don't do a parse only */
        int tag_cnt = 0;
        unsigned long mem;

        strcpy(progname, *argv++);
        argc--;
        while(argc && ((*argv)[0] == '-')){
                switch((*argv)[1]){
                        case 'v':
                                printf("%s\n",VERSION);
                                return 0;
                        case 'h':
                                printf("%s\n",USAGE);
                                printf("%s\n",HELP1);
                                printf("%s\n",HELP2);
                                printf("%s\n",HELP3);
                                printf("%s\n",HELP4);       
                                printf("%s\n",HELP5);
                                return 0;
                        case 'p':
                                active = 0;
                                break;
                        default:
                                fprintf(stderr,"%s unknown option -%c\n",
                                        progname,(*argv)[1]);
                                return 1;
                }
                argc--;argv++;
        }
        strcpy(filename,CONFIG_SYS);

        if(argc)
                strcpy(filename,*argv++);

        /* See if we can find config.sys in the usual place and open
        it for reading */

        input = fopen(filename, "r");

        if(!input){
                fprintf(stderr,"%s: cannot open %s\n",progname,filename);
                return 1;
        }

        /* Do a quick and dirty parse of input to see what sizes of various
        things we need */

        ntags = 0;
        k=0;
        while ((c = fgetc(input))!= EOF){
                k++;
                if(c == '[')ntags++; /* We don't call bracket police to make sure
                                        they're matched */
                if(c == '=')max_env++;  /* probably an overcount */
        }
        rewind(input);
        if(!active){
                printf("%d characters read from %s\n",k,filename);
                printf("%d tags, %d = statments\n",ntags,max_env);
                printf("Core Left:  %lu\n",
                mem = coreleft());
        }

        my_tags = malloc(ntags*sizeof(struct tag_struct));
        tokenbuf = malloc((k+1)*sizeof(char));
        /* BUG: these are way too big */
        current_tag = malloc((k+1)*sizeof(char));

        if((j = strlen(getenv("COMPSPEC")))!=0)
                if(k < j) k = j+1; 
        else
                if(k<strlen(COMSPEC)+1) k = strlen(COMSPEC)+1;
        shell_command_line = malloc((k+1)*sizeof(char));
        if(!my_tags || !tokenbuf ||!shell_command_line || !current_tag){
                fprintf(stderr,"%s: unable to allocate memory\n",progname);
                return 1;
        }
        tokenbuf[0]='\0';
        current_tag[0] = '\0';
        if(j) strcpy(shell_command_line,getenv("COMSPEC"));
        else strcpy(shell_command_line,COMSPEC);

        my_tags[0].name = strdup("COMMON");
        my_tags[0].shell = strdup("");
        my_tags[0].env = malloc(max_env*sizeof(char *));
        my_tags[0].nmenuitems = 0; /* No menu allowed in common block */
        my_tags[0].n_env = 0;
        tag_cnt = 1;

        /* Main parsing loop */
        while(parse_status != MY_EOF){

                p = parse(input, &parse_status);
                if(p){
                        n = strlen(p);
                        if(n){
                                if(p[0] == '['){  /* Looks like a tag */
                                for(i=0;i<n-2;i++)tokenbuf[i]=p[i+1];
                                if(p[i+1] != ']'){
                                        fprintf(stderr,"Syntax Error\n");
                                        return 1;
                                }
                                tokenbuf[i]='\0';
                                forceuc(tokenbuf);
                                if(strcmp(tokenbuf,"END")==0) break;
                                strcpy(current_tag,tokenbuf);
                                /* See if this tag already exists */
                                for(i=0;i<tag_cnt;i++)
                                        if(strcmp(my_tags[i].name,
                                                current_tag)==0)break;
                                if(i==tag_cnt){
                                        /* if new tag, initialize */
                                        my_tags[i].name = strdup(current_tag);
                                        my_tags[i].n_env = 0;
                                        my_tags[i].env = malloc(max_env*sizeof(char *));
                                        my_tags[i].shell = strdup("");
                                        my_tags[i].my_menu = malloc(ntags*sizeof( struct menuitem_struct *));
                                        my_tags[i].nmenuitems = 0;
                                        my_tags[i].menudefault = 1;
                                        my_tags[i].menutimeout = -1;
                                        my_tags[i].menuforeground = 15;
                                        my_tags[i].menubackground = 1;
                                        tag_cnt++;
                                }
                                j_t = i;
                                }
                                else { /* Here we process the various non-tag
                                tokens that matter to us. Current tag # is 
                                j_t */
                                        forceuc(p);
                                        if(strcmp("REM",p)==0){
                                                while(parse_status != MY_EOL)
                                                        parse(input,&parse_status);
                                                continue;
                                        }
                                        if(strcmp("SHELL",p)==0){
                                                tokenbuf[0]='\0';
                                                while(parse_status != MY_EOL){
                                                p = parse(input,&parse_status);
                                                strcat(tokenbuf,p);
                                                strcat(tokenbuf," ");
                                                }
                                                my_tags[j_t].shell = strdup(tokenbuf);
                                                continue;
                                        }
                                        if(strcmp(p,"SET")==0){
                                                tokenbuf[0]='\0';
                                                p = parse(input,&parse_status);
                                                strcpy(tokenbuf,p);
                                                strcat(tokenbuf,"=");
                                                p = parse(input,&parse_status);
                                                
                                                strcat(tokenbuf,p);
                                                n = my_tags[j_t].n_env;
                                                my_tags[j_t].env[n] 
                                                        = strdup(tokenbuf);
                                                my_tags[j_t].n_env++;
                                                continue;
                                        }
                                        if(strcmp(p,"INCLUDE")==0){
                                                p = parse(input,&parse_status);
                                                for(i=0;i<ntags;i++)
                                                        if(strcmp(p,my_tags[i].name)==0)break;
                                                if(i<ntags){
                                                        for(j=0;j<my_tags[i].n_env;j++)
                                                                my_tags[j_t].env[my_tags[j_t].n_env++]
                                                                        = my_tags[i].env[j];
                                                        if(strlen(my_tags[i].shell))
                                                                my_tags[j_t].shell = my_tags[i].shell;
                                                        
                                                }
                                                continue;
                                        }
                                        if(strcmp(p,"MENUITEM")==0){
                                                if(j_t == 0) continue; 
                                                n = my_tags[j_t].nmenuitems++;
                                                my_tags[j_t].my_menu[n] = alloc_menuitem();
                                                (my_tags[j_t].my_menu[n])->n = n;
                                                p = parse(input,&parse_status);
                                                (my_tags[j_t].my_menu[n])->tag = strdup(p);
                                                tokenbuf[0]='\0';
                                                while(parse_status != MY_EOL){
                                                  p = parse(input,&parse_status);
                                                  strcat(tokenbuf,p);
                                                  strcat(tokenbuf," ");
                                                }
                                                (my_tags[j_t].my_menu[n])->text = strdup(tokenbuf);
                                                continue;
                                        }
                                        if(strcmp(p,"MENUCOLOR")==0){
                                                p = parse(input,&parse_status);
                                                my_tags[j_t].menuforeground =
                                                        atoi(p);
                                                p = parse(input,&parse_status);
                                                my_tags[j_t].menubackground = 
                                                atoi(p);
                                                continue;
                                        }
                                        if(strcmp(p,"MENUDEFAULT")==0){
                                                if(j_t == 0) continue;
                                                p = parse(input,&parse_status);
                                                for(i=0;i<my_tags[j_t].nmenuitems;i++)
                                                        if(strcmp(p,(my_tags[j_t].my_menu[i])->tag)==0)break;
                                                if(i<my_tags[j_t].nmenuitems)
                                                        my_tags[j_t].menudefault = ((my_tags[j_t].my_menu[i])->n)+1;
                                                if(parse_status != MY_EOL){
                                                        p = parse(input,&parse_status);
                                                        my_tags[j_t].menutimeout = atoi(p);
                                                }
                                                continue;
                                        }
                                 }  
                        }
                }
        }
        fclose(input);

        if(!active){
                printf("core used by program = %lu\n", mem - coreleft());
                return 0;
        }
        

        /* Add common block environment strings */
        for(j=0;j<my_tags[0].n_env;j++)
                putenv(my_tags[0].env[j]);
        /* Override default shell command line if there was a shell=
           in the common block */
        if(strlen(my_tags[0].shell))
                strcpy(shell_command_line, my_tags[0].shell);

        /* Search for the MENU TAG. If it is not there, then we have
        nothing further to do */
        for(i=0;i<ntags;i++)
                if(strcmp(my_tags[i].name,"MENU")==0)break;
        if(i<ntags){
                while((j = do_tag(i))!=0){
                        for(k=0;k<ntags;k++)
                                if(strcmp(my_tags[k].name,(my_tags[i].my_menu[j-1])->tag)==0)break;
                        if(k<ntags)i=k;
                        else break;
                }

        /* do_tag does putenv's, overrides the shell choice, [...?] and
        presents a menu if it is menu tag. It returns the user's choice
        to recursively allow for submenus. Returns 0 when there is nothing
        more to do. */
        }
#if 1
        /* Clean up */
        free(tokenbuf);
        free(current_tag);
        for(i=0;i<ntags;i++){
                free(my_tags[i].shell);
                /* freeing environment strings seems to leave then corrupted 
                for(j=0;j<my_tags[i].n_env;j++)
                        free(my_tags[i].env[j]); */
                free(my_tags[i].env);
                for(j=0;j<my_tags[i].nmenuitems;j++){
                        free((my_tags[i].my_menu[j])->tag);
                        free((my_tags[i].my_menu[j])->text);
                        free(my_tags[i].my_menu[j]);
                }
                free(my_tags[i].name);
        }
        free(my_tags);
#endif
        /* And, take it away .... */
/*        system(shell_command_line); */

printf("shell_command_line = %s\n", shell_command_line);
        free(shell_command_line); /* Unreachable ? */

        return 0;
}

/* Enact the tag whose number is passed in. First do any putenv's it needs 
done, then override the current shell command line if there is a shell=. If
it is a menu tag, present the user with a menu and return their choice. 
Otherwise return 0 */

int do_tag(int j)
{
          struct tag_struct *t;
          int i;

          if((j<1)||(j>=ntags))return 0;

          t = &(my_tags[j]);  
          for(i=0;i<t->n_env;i++)
                putenv((t->env)[i]);
          if(strlen(t->shell))
                strcpy(shell_command_line,t->shell);

          if(t->nmenuitems)
                return menu_text(t->my_menu,t->nmenuitems,
                        t->menudefault,t->menutimeout,t->menuforeground,
                        t->menubackground);

          else return 0;
}

/* forceuc: make sure all characters in a string are upper case */

void forceuc(char *s)
{
        int n,i;
        if(s){
                n = strlen(s);
                for(i=0;i<n;i++)
                        if(islower(s[i]))
                                s[i] = toupper(s[i]);
        }
}
#ifdef NO_STRDUP
char *strdup(char *s)
{
        char *p;
        if(!s)return NULL;
        p = malloc(strlen(s)+1);
        if(!p)return NULL;
        strcpy(p,s);
        return p;
}
#endif


 


